/*
Licensed Materials - Property of IBM
5725-C94, 5725-C95, 5725-C96, 5655-Y02
(C) Copyright IBM Corporation 2012, 2016. All Rights Reserved.
This sample program is provided AS IS and may be used, executed, copied and modified without royalty payment by customer
(a) for its own instruction and study, (b) in order to develop applications designed to run with an IBM product,
either for customer's own internal use or for redistribution by customer, as part of such an application, in customer's
own products.
*/

package com.ibm.bpm.integration.filenet;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.resource.spi.security.PasswordCredential;
import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.AuthScope;

import com.ibm.websphere.security.NotImplementedException;
import com.ibm.wsspi.security.auth.callback.Constants;
import com.ibm.wsspi.security.auth.callback.WSMappingCallbackHandlerFactory;

public class BPMClientConfig {

    public static final String BPM_SERVER_URI_PROPERTY = "bpm.server.uri";
    public static final String BPM_SERVER_CONTEXTROOT_PROPERTY = "bpm.server.contextRoot";
    public static final String BPM_SERVER_PASSWORD_PROPERTY = "bpm.server.password";
    public static final String BPM_SERVER_USERNAME_PROPERTY = "bpm.server.username";
    public static final String BPM_SERVER_AUTH_ALIAS_PROPERTY = "bpm.server.authalias";

    private static final String CLASS_NAME = BPMEventHandler.class.getName();
    private static final Logger LOGGER = Logger.getLogger(CLASS_NAME);

    private URI targetServer = null;
    private String contextRoot = null;
    private String userName = "UNAUTHENTICATED";
    private String password = null;
    private String authAlias = null;
    private final List<HttpClient> httpClientCache = new LinkedList<HttpClient>();
    private final Set<HttpClient> reservedClientCache = new HashSet<HttpClient>();

    public BPMClientConfig(Properties props) {

        try {
            this.targetServer = new URI(props.getProperty(BPM_SERVER_URI_PROPERTY, "http://localhost:9080"));
        } catch (URISyntaxException e) {
            throw new IllegalArgumentException(e);
        }

        this.contextRoot = props.getProperty(BPM_SERVER_CONTEXTROOT_PROPERTY, "/rest/bpm/wle/");
        if(!this.contextRoot.startsWith("/")) {
            this.contextRoot = "/" + this.contextRoot;
        }
        if(!this.contextRoot.endsWith("/")) {
            this.contextRoot = this.contextRoot + "/";
        }

        this.userName = props.getProperty(BPM_SERVER_USERNAME_PROPERTY, "UNAUTHENTICATED");
        this.password = props.getProperty(BPM_SERVER_PASSWORD_PROPERTY, "");
        this.authAlias = props.getProperty(BPM_SERVER_AUTH_ALIAS_PROPERTY);

        if (this.authAlias != null) {
            PasswordCredential pwc = lookupCredentials();
            if(pwc != null) {
                this.userName = pwc.getUserName();
                this.password = String.valueOf(pwc.getPassword());
            } else {
                this.userName = props.getProperty(BPM_SERVER_USERNAME_PROPERTY, "UNAUTHENTICATED");
                this.password = props.getProperty(BPM_SERVER_PASSWORD_PROPERTY, "");
            }
        }
    }

    public void update(Properties properties) {
        boolean dirty = false;

        try {
            URI targetServer = new URI(properties.getProperty(BPM_SERVER_URI_PROPERTY, "http://localhost:9080"));

            if(!targetServer.equals(this.targetServer)) {
                this.targetServer = targetServer;
                dirty = true;
            }
        } catch (URISyntaxException e) {
            throw new IllegalArgumentException(e);
        }

        String contextRoot = properties.getProperty(BPM_SERVER_CONTEXTROOT_PROPERTY, "/rest/bpm/wle/");
        if(!contextRoot.equals(this.contextRoot)) {
            this.contextRoot = contextRoot;
            dirty = true;
        }

        String authAlias = properties.getProperty(BPM_SERVER_AUTH_ALIAS_PROPERTY);
        if(authAlias == null) {
            if(this.authAlias != null) {
                this.authAlias = null;
                dirty = true;
            }

            String username = properties.getProperty(BPM_SERVER_USERNAME_PROPERTY, "UNAUTHENTICATED");
            if(!username.equals(this.userName)) {
                this.userName = username;
                dirty = true;
            }

            String password = properties.getProperty(BPM_SERVER_PASSWORD_PROPERTY, "");
            if(!password.equals(this.password)) {
                this.password = password;
                dirty = true;
            }
        } else {
            if(this.authAlias == null || !this.authAlias.equals(authAlias)) {
                this.authAlias = authAlias;
                dirty = true;
            }

            PasswordCredential pwc = lookupCredentials();
            final String username;
            final String password;
            if(pwc != null) {
                username = pwc.getUserName();
                password = String.valueOf(pwc.getPassword());
            } else {
                username = properties.getProperty(BPM_SERVER_USERNAME_PROPERTY, "UNAUTHENTICATED");
                password = properties.getProperty(BPM_SERVER_PASSWORD_PROPERTY, "");
            }

            if(!username.equals(this.userName)) {
                this.userName = username;
                dirty = true;
            }
            if(!password.equals(this.password)) {
                this.password = password;
                dirty = true;
            }
        }

        if(dirty) {
            this.httpClientCache.clear();
            this.reservedClientCache.clear();
        }
    }

    /**
     * Reserves a HTTP client. This method reuses a free client from the cache or creates a new one.
     *
     * @return a HttpClient instance
     */
    public synchronized HttpClient getHttpClient() {
        final String methodName = "getHttpClient";
        if (LOGGER.isLoggable(Level.FINEST)) {
            LOGGER.entering(CLASS_NAME, methodName);
        }

        if(!this.httpClientCache.isEmpty()) {
            HttpClient httpClient = this.httpClientCache.remove(0);
            this.reservedClientCache.add(httpClient);

            if (LOGGER.isLoggable(Level.FINEST)) {
                LOGGER.logp(Level.FINEST, CLASS_NAME, methodName, "Using HttpClient from cache.");
                LOGGER.exiting(CLASS_NAME, methodName, httpClient);
            }
            return httpClient;
        } else {
            AuthScope authScope = new AuthScope(this.targetServer.getHost(), this.targetServer.getPort());
            UsernamePasswordCredentials creds = new UsernamePasswordCredentials(this.userName, this.password);

            HttpClient httpClient = new HttpClient();
            httpClient.getParams().setAuthenticationPreemptive(true);
            httpClient.getState().setCredentials(authScope, creds);

            this.reservedClientCache.add(httpClient);

            if (LOGGER.isLoggable(Level.FINEST)) {
                LOGGER.logp(Level.FINEST, CLASS_NAME, methodName, "Created new HttpClient.");
                LOGGER.exiting(CLASS_NAME, methodName, httpClient);
            }
            return httpClient;
        }
    }

    /**
     * Releases a HTTP client to allow reusing it for other events.
     *
     * @param httpClient the HttpClient instance
     */
    public synchronized void releaseHttpClient(HttpClient httpClient) {
        final String methodName = "releaseHttpClient";
        if (LOGGER.isLoggable(Level.FINEST)) {
            LOGGER.entering(CLASS_NAME, methodName, httpClient);
        }

        // only add the object to the cache if it is currently reserved
        if(this.reservedClientCache.remove(httpClient)) {
            this.httpClientCache.add(httpClient);
        }

        if (LOGGER.isLoggable(Level.FINEST)) {
            LOGGER.exiting(CLASS_NAME, methodName);
        }
    }

    @SuppressWarnings({ "rawtypes", "unchecked" })
    private PasswordCredential lookupCredentials() {
        String methodName = "lookupCredentials";
        if (LOGGER.isLoggable(Level.FINEST)) {
            LOGGER.entering(CLASS_NAME, methodName);
        }
        try {
            Map map = new HashMap();
            map.put(Constants.MAPPING_ALIAS, authAlias);
            // create a callback handler with the specified property (to find the alias) and null for the managed connection factory (MCF) since we don't need it
            CallbackHandler handler = WSMappingCallbackHandlerFactory.getInstance().getCallbackHandler(map, null);

            Subject subject = new Subject();
            LoginContext lc = new LoginContext("DefaultPrincipalMapping", subject, handler);
            if (LOGGER.isLoggable(Level.FINEST)) LOGGER.log(Level.FINEST, "trying to login via authentication alias: {0}", authAlias );
            lc.login();
            subject = lc.getSubject();
            Set<PasswordCredential> pwdCredentialSet = subject.getPrivateCredentials(PasswordCredential.class);
            PasswordCredential pwCred = pwdCredentialSet.iterator().next();

            lc.logout();
            if (LOGGER.isLoggable(Level.FINEST)) {
                LOGGER.exiting(CLASS_NAME, methodName, pwCred);
            }
            return pwCred;
        } catch (LoginException e) {
            if(LOGGER.isLoggable(Level.FINER)) {
                LOGGER.logp(Level.FINER, CLASS_NAME, methodName, "Using authentication alias failed.", e);
            }

            return null;
        } catch (NotImplementedException e) {
            throw new AssertionError();
        }
    }

    /**
     * Set the password to be used for authentication
     *
     * @param password the password to set
     */
    public void setPassword(String password) {
        this.password = password;
    }

    /**
     * Get the context root of the BPM REST API
     *
     * @return a context root that will start and end with a slash
     */
    protected String getContextRoot() {
        return contextRoot;
    }

    /**
     * Get the URI of the target server
     *
     * @return a URI
     */
    protected URI getTargetServer() {
        return targetServer;
    }

    /**
     * Get the user name to be used for authentication
     *
     * @return a user name
     */
    protected String getUserName() {
        return userName;
    }

    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer("TargetServer: '").append(this.targetServer.toString());
        buf.append("', contextRoot: '").append(contextRoot);
        if (authAlias != null) {
            buf.append("', AuthenticationAlias: '").append(authAlias);
        }
        buf.append("' , username: '").append(this.userName).append("' , password: '********' ");
        buf.append(" objecthash: ").append(super.toString());
        return buf.toString();
    }
}
