/*
Licensed Materials - Property of IBM
5725-C94, 5725-C95, 5725-C96, 5655-Y02
(C) Copyright IBM Corporation 2012, 2016. All Rights Reserved.
This sample program is provided AS IS and may be used, executed, copied and modified without royalty payment by customer
(a) for its own instruction and study, (b) in order to develop applications designed to run with an IBM product,
either for customer's own internal use or for redistribution by customer, as part of such an application, in customer's
own products.
*/

package com.ibm.bpm.integration.filenet;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.filenet.api.constants.ClassNames;
import com.filenet.api.constants.PropertyNames;
import com.filenet.api.constants.VersionStatus;
import com.filenet.api.core.Document;
import com.filenet.api.core.Factory;
import com.filenet.api.core.Folder;
import com.filenet.api.core.IndependentObject;
import com.filenet.api.core.ObjectStore;
import com.filenet.api.core.ReferentialContainmentRelationship;
import com.filenet.api.engine.EventActionHandler;
import com.filenet.api.events.CancelCheckoutEvent;
import com.filenet.api.events.ChangeClassEvent;
import com.filenet.api.events.ChangeStateEvent;
import com.filenet.api.events.CheckinEvent;
import com.filenet.api.events.CheckoutEvent;
import com.filenet.api.events.ClassifyCompleteEvent;
import com.filenet.api.events.CreationEvent;
import com.filenet.api.events.DeletionEvent;
import com.filenet.api.events.DemoteVersionEvent;
import com.filenet.api.events.FileEvent;
import com.filenet.api.events.FreezeEvent;
import com.filenet.api.events.LockEvent;
import com.filenet.api.events.ObjectChangeEvent;
import com.filenet.api.events.PromoteVersionEvent;
import com.filenet.api.events.PublishCompleteEvent;
import com.filenet.api.events.PublishRequestEvent;
import com.filenet.api.events.Subscription;
import com.filenet.api.events.UnfileEvent;
import com.filenet.api.events.UnlockEvent;
import com.filenet.api.events.UpdateEvent;
import com.filenet.api.events.UpdateSecurityEvent;
import com.filenet.api.exception.EngineRuntimeException;
import com.filenet.api.exception.ExceptionCode;
import com.filenet.api.property.Property;
import com.filenet.api.property.PropertyFilter;
import com.filenet.api.util.Id;

public class BPMEventHandler implements EventActionHandler {

    private static final String CLASS_NAME = BPMEventHandler.class.getName();
    private static final Logger LOGGER = Logger.getLogger(CLASS_NAME);

    private static class CacheEntry {

        private BPMClientConfig clientConfig;
        private volatile Id documentId;
        private volatile LongReference expires;

        private CacheEntry(BPMClientConfig clientConfig, Id documentId, LongReference expires) {
            this.clientConfig = clientConfig;
            this.documentId = documentId;
            this.expires = expires;
        }
    }

    private static class LongReference {

        private volatile long l;

        private LongReference(long l) {
            this.l = l;
        }

    }

    private static final Map<Id, CacheEntry> CONFIGURATION_CACHE = new HashMap<Id, CacheEntry>();

    private static final String CMIS_DOCUMENT_SYMBOLIC_NAME = "cmis:document";
    private static final String CMIS_FOLDER_SYMBOLIC_NAME = "cmis:folder";

    private static final PropertyFilter PF_DOCUMENT = new PropertyFilter();
    private static final PropertyFilter PF_OBJECT_STORE_ID = new PropertyFilter();
    private static final PropertyFilter PF_USER_STRING = new PropertyFilter();

    static {
        PF_DOCUMENT.addIncludeProperty(1, null, Boolean.FALSE,
            PropertyNames.CONTENT_ELEMENTS + " " +
            PropertyNames.CURRENT_VERSION + " " +
            PropertyNames.ID + " " +
            PropertyNames.IS_CURRENT_VERSION, null);
        PF_OBJECT_STORE_ID.addIncludeProperty(0, null, Boolean.FALSE, PropertyNames.OBJECT_STORE_ID, null);
        PF_USER_STRING.addIncludeProperty(0, null, Boolean.FALSE, PropertyNames.USER_STRING, null);
    }

    private BPMEventType getBPMEventType(ObjectChangeEvent event) {
        String methodName = "getBPMEventType";
        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.entering(CLASS_NAME, methodName, event);

        BPMEventType eventType = null;
        if (event instanceof CancelCheckoutEvent) {
            eventType = BPMEventType.CheckOutCanceled;
        } else if (event instanceof ChangeClassEvent) {
            eventType = BPMEventType.ClassChanged;
        } else if (event instanceof ChangeStateEvent) {
            eventType = BPMEventType.StateChanged;
        } else if (event instanceof CheckinEvent) {
            eventType = BPMEventType.CheckedIn;
        } else if (event instanceof CheckoutEvent) {
            eventType = BPMEventType.CheckedOut;
        } else if (event instanceof ClassifyCompleteEvent) {
            eventType = BPMEventType.ClassifyCompleted;
        } else if (event instanceof CreationEvent) {
            eventType = BPMEventType.Created;
        } else if (event instanceof DeletionEvent) {
            eventType = BPMEventType.Deleted;
        } else if (event instanceof DemoteVersionEvent) {
            eventType = BPMEventType.VersionDemoted;
        } else if (event instanceof FileEvent) {
            eventType = BPMEventType.Filed;
        } else if (event instanceof FreezeEvent) {
            eventType = BPMEventType.Frozen;
        } else if (event instanceof LockEvent) {
            eventType = BPMEventType.Locked;
        } else if (event instanceof PromoteVersionEvent) {
            eventType = BPMEventType.VersionPromoted;
        } else if (event instanceof PublishCompleteEvent) {
            eventType = BPMEventType.PublishCompleted;
        } else if (event instanceof PublishRequestEvent) {
            eventType = BPMEventType.PublishRequested;
        } else if (event instanceof UnfileEvent) {
            eventType = BPMEventType.Unfiled;
        } else if (event instanceof UnlockEvent) {
            eventType = BPMEventType.Unlocked;
        } else if (event instanceof UpdateEvent) {
            eventType = BPMEventType.Updated;
        } else if (event instanceof UpdateSecurityEvent) {
            eventType = BPMEventType.SecurityUpdated;
        } else {
            if (LOGGER.isLoggable(Level.WARNING)) LOGGER.warning("The ObjectChangeEvent named " + event.get_Name() + " has been ignored as there is no corresponding BPM ECMEvent it can be mapped to.");
        }
        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.exiting(CLASS_NAME, methodName, eventType);
        return eventType;
    }

    private ECMEventPoster getEventPoster(BPMClientConfig clientConfig) {
        String methodName = "getEventPoster";
        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.entering(CLASS_NAME, methodName, clientConfig);

        ECMEventPoster eventPoster = new ECMEventPoster(clientConfig);

        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.exiting(CLASS_NAME, methodName, eventPoster);
        return eventPoster;
    }

    private String getEventSourceId(ObjectStore objectStore) {
        String methodName = "getEventSourceId";
        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.entering(CLASS_NAME, methodName, objectStore);

        com.filenet.api.property.Properties osProperties = objectStore.getProperties();
        Property objectStoreIdProperty = osProperties.find(PropertyNames.OBJECT_STORE_ID);
        if(objectStoreIdProperty == null) {
            objectStore.fetchProperties(PF_OBJECT_STORE_ID);
            objectStoreIdProperty = osProperties.get(PropertyNames.OBJECT_STORE_ID);
        }
        String eventSourceId = objectStoreIdProperty.getIdValue().toString();

        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.exiting(CLASS_NAME, methodName, eventSourceId);
        return eventSourceId;
    }

    private String getRelatedObjectId(ObjectChangeEvent event) {
        String methodName = "getRelatedObjectId";
        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.entering(CLASS_NAME, methodName, event);

        String relatedObjectId = null;

        if(event instanceof FileEvent || event instanceof UnfileEvent) {
            IndependentObject source = event.get_SourceObject();
            if(source instanceof ReferentialContainmentRelationship) {
                // Document filed into folder / unfiled from folder
                ReferentialContainmentRelationship rcr = (ReferentialContainmentRelationship)source;
                IndependentObject head = rcr.get_Head();
                // Head is null if the unfile happens as part of a document deletion
                if(head != null) {
                    String objectId = head.getProperties().getIdValue(PropertyNames.ID).toString();

                    StringBuffer cmisFormOID = new StringBuffer(40);
                    cmisFormOID.append("idd_");
                    cmisFormOID.append(objectId.substring(1, objectId.length()-1)); //strip curly brackets {}

                    relatedObjectId = cmisFormOID.toString();
                }
            } else if(source instanceof Folder) {
                // Folder added to folder / removed from folder
                Folder folder = (Folder)source;
                String objectId = folder.get_Id().toString();

                StringBuffer cmisFormOID = new StringBuffer(40);
                cmisFormOID.append("idf_");
                cmisFormOID.append(objectId.substring(1, objectId.length()-1)); //strip curly brackets {}

                relatedObjectId = cmisFormOID.toString();
            }
        }

        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.exiting(CLASS_NAME, methodName, relatedObjectId);
        return relatedObjectId;
    }

    private String getRelatedObjectTypeId(ObjectChangeEvent event) {
        String methodName = "getRelatedObjectTypeId";
        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.entering(CLASS_NAME, methodName, event);

        String relatedObjectTypeId = null;

        if(event instanceof FileEvent || event instanceof UnfileEvent) {
            IndependentObject source = event.get_SourceObject();
            if(source instanceof ReferentialContainmentRelationship) {
                // Document filed into folder / unfiled from folder
                ReferentialContainmentRelationship rcr = (ReferentialContainmentRelationship)source;
                IndependentObject head = rcr.get_Head();
                // Head is null if the unfile happens as part of a document deletion
                if(head != null) {
                    relatedObjectTypeId = rcr.get_Head().getClassName();
                }
            } else if(source instanceof Folder) {
                // Folder added to folder / removed from folder
                relatedObjectTypeId = source.getClassName();
            }

            // translate Document to cmis:document and Folder to cmis:folder
            if(relatedObjectTypeId != null) {
                if(relatedObjectTypeId.equals(ClassNames.DOCUMENT)) {
                    relatedObjectTypeId = CMIS_DOCUMENT_SYMBOLIC_NAME;
                } else if(relatedObjectTypeId.equals(ClassNames.FOLDER)) {
                    relatedObjectTypeId = CMIS_FOLDER_SYMBOLIC_NAME;
                }
            }
        }

        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.exiting(CLASS_NAME, methodName, relatedObjectTypeId);
        return relatedObjectTypeId;
    }

    private String getObjectTypeId(ObjectChangeEvent event) {
        String methodName = "getObjectTypeId";
        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.entering(CLASS_NAME, methodName, event);

        String objectTypeId = Factory.MetadataCache.getDefaultInstance().getClassDescription(event.getObjectStore(), event.get_SourceClassId().toString()).get_SymbolicName();

        // translate Document to cmis:document and Folder to cmis:folder
        if (objectTypeId.equals(ClassNames.DOCUMENT)) {
            objectTypeId = CMIS_DOCUMENT_SYMBOLIC_NAME;
        } else if (objectTypeId.equals(ClassNames.FOLDER)) {
            objectTypeId = CMIS_FOLDER_SYMBOLIC_NAME;
        }

        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.exiting(CLASS_NAME, methodName, objectTypeId);
        return objectTypeId;
    }

    private String getObjectId(ObjectChangeEvent event) {
        String methodName = "getObjectCMISId";
        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.entering(CLASS_NAME, methodName, event);

        IndependentObject sourceObject = event.get_SourceObject();
        String objectId = event.get_SourceObjectId().toString();
        StringBuffer cmisFormOID = new StringBuffer(40);

        //prefix with 'idf_' for objects with base type folder, or with 'idd_' for objects with base type document
        // For the File and Unfile event the sourceObject is a DynamicReferentialContainmentRelationship object
        if (sourceObject instanceof Folder || sourceObject instanceof ReferentialContainmentRelationship) {
            cmisFormOID.append("idf_");
        }
        else if (sourceObject instanceof Document) {
            cmisFormOID.append("idd_");
        }

        cmisFormOID.append(objectId.substring(1, objectId.length()-1)); //strip curly brackets {}

        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.exiting(CLASS_NAME, methodName, cmisFormOID.toString());
        return cmisFormOID.toString();
    }

    @Override
    public void onEvent(ObjectChangeEvent event, Id subId) throws EngineRuntimeException {
        String methodName = "onEvent";
        if (LOGGER.isLoggable(Level.FINEST)) LOGGER.entering(CLASS_NAME, methodName, new Object[]{event, subId});

        try {
            if(event instanceof CreationEvent && !BPMEventHandlerConfiguration.isSendCreatedEventOnCheckOut() && event.get_SourceObject() != null && event.get_SourceObject() instanceof Document) {
                // Make sure that only the "real" document creation is handled and not a checkout of a document that creates another reservation document
                Document document = (Document)event.get_SourceObject();
                if(document.get_IsVersioningEnabled()) {
                    boolean isDocumentCreation =
                            Boolean.TRUE.equals(document.get_IsReserved()) &&
                            document.get_MajorVersionNumber() == 0 &&
                            document.get_MinorVersionNumber() == 1 &&
                            VersionStatus.RESERVATION.equals(document.get_VersionStatus());

                    if(!isDocumentCreation) {
                        if(LOGGER.isLoggable(Level.FINEST)) {
                            LOGGER.logp(Level.FINEST, CLASS_NAME, methodName, "This event is not triggered by a document creation and being ignored. Is Reserved: " + document.get_IsReserved() + ". Major Version Number: " + document.get_MajorVersionNumber() + ". Minor Version Number: " + document.get_MinorVersionNumber() + ". Version Status: " + document.get_VersionStatus());
                        }
                        if (LOGGER.isLoggable(Level.FINER)) {
                            LOGGER.exiting(CLASS_NAME, methodName);
                        }
                        return;
                    }
                }
            } // end if(event instanceof CreationEvent && !BPMEventHandlerConfiguration.isSendCreatedEventOnCheckOut() && event.get_SourceObject() != null && event.get_SourceObject() instanceof Document)

            ObjectStore os = event.getObjectStore();

            // Check if we have cached the properties
            long now = System.currentTimeMillis();
            CacheEntry cache = CONFIGURATION_CACHE.get(subId);
            if(cache == null) {
                Document configDocument = this.getConfigDocument(os, subId);

                // look if any entry in the cache has the same documentId
                BPMClientConfig clientConfig = null;
                LongReference expires = null;
                synchronized (CONFIGURATION_CACHE) {
                    for(CacheEntry candidate : CONFIGURATION_CACHE.values()) {
                        synchronized (candidate) {
                            if(candidate.documentId.equals(configDocument.get_Id())) {
                                clientConfig = candidate.clientConfig;
                                expires = candidate.expires;
                                break;
                            }
                        }
                    }
                }

                if(clientConfig == null) {
                    Properties properties = new Properties();
                    properties.load(configDocument.accessContentStream(0));
                    clientConfig = new BPMClientConfig(properties);
                    expires = new LongReference(now + BPMEventHandlerConfiguration.getConfigurationCacheExpirationTime());
                }

                cache = new CacheEntry(clientConfig, configDocument.get_Id(), expires);
                synchronized (CONFIGURATION_CACHE) {
                    CONFIGURATION_CACHE.put(subId, cache);
                }
            }

            if(cache.expires.l > now) {
                synchronized (cache) {
                    if(cache.expires.l > now) {
                        Document configDocument = this.getConfigDocument(os, subId);
                        Properties properties = new Properties();
                        properties.load(configDocument.accessContentStream(0));

                        if(!cache.documentId.equals(configDocument.get_Id())) {
                            cache.clientConfig = new BPMClientConfig(properties);
                            cache.documentId = configDocument.get_Id();
                        } else {
                            cache.clientConfig.update(properties);
                        }
                        cache.expires.l = now + BPMEventHandlerConfiguration.getConfigurationCacheExpirationTime();
                    }
                }
            }

            ECMEvent ecmEvent = new ECMEvent(getEventSourceId(os), getBPMEventType(event), getObjectTypeId(event), getObjectId(event), getRelatedObjectTypeId(event), getRelatedObjectId(event));
            int rc = getEventPoster(cache.clientConfig).postEvent(ecmEvent);

            if (LOGGER.isLoggable(Level.FINEST)) LOGGER.exiting(CLASS_NAME, methodName, rc);

         } catch (Exception e) {
            if (LOGGER.isLoggable(Level.FINEST)) {
                LOGGER.log(Level.FINE, "Exception when trying to post the event to BPM the ExceptionMessage was: {0}", e.getMessage());
                LOGGER.log(Level.FINEST,"StackTrace: ", e);
            }
            throw new EngineRuntimeException(e, ExceptionCode.EVENT_HANDLER_THREW, null);
        }
    }

    private Document getConfigDocument(ObjectStore os, Id subscriptionId) {
        final String METHOD_NAME = "getConfigDocument";
        if(LOGGER.isLoggable(Level.FINER)) {
            LOGGER.entering(CLASS_NAME, METHOD_NAME, new Object[]{os, subscriptionId});
        }

        // get the user string of the subscription -> our connection document identifier
        Subscription subscription = Factory.Subscription.fetchInstance(os, subscriptionId, PF_USER_STRING);
        String userString = subscription.get_UserString();

        if(LOGGER.isLoggable(Level.FINEST)) {
            LOGGER.logp(Level.FINEST, CLASS_NAME, METHOD_NAME, "Calculating event handler configuration based on UserString {0}", userString );
        }

        if(userString == null) {
            IllegalArgumentException e = new IllegalArgumentException("UserString == null. Check this property on the subscription with the '" + subscriptionId + "' identifier.");
            if(LOGGER.isLoggable(Level.FINER)) {
                LOGGER.throwing(CLASS_NAME, METHOD_NAME, e);
            }
            throw e;
        }

        Document configDoc = Factory.Document.fetchInstance(os, userString, PF_DOCUMENT);

        if(!configDoc.get_IsCurrentVersion()) {
            if(LOGGER.isLoggable(Level.FINEST)) {
                LOGGER.log(Level.FINEST, "there's a newer version of the properties in the ObjectStore, trying to get the current version");
            }
            configDoc = (Document)configDoc.get_CurrentVersion();
        }

        if(LOGGER.isLoggable(Level.FINER)) {
            LOGGER.exiting(CLASS_NAME, METHOD_NAME, configDoc);
        }
        return configDoc;
    }
}